# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import io
import re
import tarfile
try:
    import lzma
except ImportError:
    from backports import lzma

from debtools.ar import ArFile

__author__ = 'Matthieu Gallet'


def parse_control_data(control_data, continue_line=' ', split=': ', skip_after_blank=False):
    offset = len(continue_line)
    result_data = {}
    key, value = None, None
    description = ''
    add_to_description = False
    for line in control_data.splitlines():
        if not line.split() and skip_after_blank:
            add_to_description = True
        if add_to_description:
            description += "\n"
            description += line
            continue
        if not line or line[0:offset] == continue_line:
            if key is not None:
                value += "\n"
                value += line[offset:]
        else:
            if key is not None:
                result_data[key] = value
            key, value = line.split(split, 1)
            value = value.lstrip()
    if key is not None:
        result_data[key] = value
    if add_to_description:
        result_data['description'] = description
    return result_data


def get_subfile(ar_file, name_regexp='control.tar.'):
    """Find the file whose names matches the given regexp
    :param ar_file:
    :type ar_file: :class:`ArFile`
    :param name_regexp:
    :type name_regexp: :class:`str`
    :return: the tuple (file descriptor, name)
    :rtype: :class:`tuple`
    """
    for name in ar_file.getnames():
        if re.match(name_regexp, name):
            return ar_file.extractfile(name), name
    return None, None


def get_control_data(filename):
    """ Extract control data from a `.deb` file
    A `.deb` is an `.ar` file that contains `control.tar.XXX`, that contains a `control` file.

    :param filename: complete filepath of a `.deb` file
    :type filename: :class:`str`
    :return:
    :rtype: :class:`dict`
    """
    # open the four files (.deb, .ar, control.tar.XXX, control)
    deb_file = open(filename, mode='rb')
    ar_file = ArFile(filename, mode='r', fileobj=deb_file)
    control_file, control_file_name = get_subfile(ar_file, '^control\.tar\..*$')
    mode = 'r:*'
    if control_file_name.endswith('.xz') or control_file_name.endswith('.lzma'):
        # special case when lzma is used from backport (Python 3.2, 2.7)
        control_file_content = control_file.read()
        control_file_content_uncompressed = lzma.decompress(control_file_content)
        control_file.close()
        control_file = io.BytesIO(control_file_content_uncompressed)
        mode = 'r'
    tar_file = tarfile.open(name='control', mode=mode, fileobj=control_file)
    control_data = tar_file.extractfile('./control')
    # we got the data!
    control_data_value = control_data.read().decode('utf-8')
    # ok, we close the previous file descriptors
    control_data.close()
    tar_file.close()
    ar_file.close()
    deb_file.close()
    # here we are
    parsed_data = parse_control_data(control_data_value)
    return parsed_data
